@@ -35,6 +35,26 @@
 #include "sstp-private.h"
 #include "sstp-client.h"
 
+/*! OpenSSL cipher suites
+ * 
+ * https://wiki.mozilla.org/Security/Server_Side_TLS
+ * Intermediate compatibility (default), as SSTP appeared only in WinSrv2k8
+ */
+
+static const char* const sstp_client_ssl_ciphers =
+    "ECDHE-ECDSA-CHACHA20-POLY1305:ECDHE-RSA-CHACHA20-POLY1305" \
+    ":ECDHE-ECDSA-AES128-GCM-SHA256:ECDHE-RSA-AES128-GCM-SHA256" \
+    ":ECDHE-ECDSA-AES256-GCM-SHA384:ECDHE-RSA-AES256-GCM-SHA384" \
+    ":DHE-RSA-AES128-GCM-SHA256:DHE-RSA-AES256-GCM-SHA384" \
+    ":ECDHE-ECDSA-AES128-SHA256:ECDHE-RSA-AES128-SHA256:ECDHE-ECDSA-AES128-SHA" \
+    ":ECDHE-RSA-AES256-SHA384:ECDHE-RSA-AES128-SHA:ECDHE-ECDSA-AES256-SHA384" \
+    ":ECDHE-ECDSA-AES256-SHA:ECDHE-RSA-AES256-SHA:DHE-RSA-AES128-SHA256" \
+    ":DHE-RSA-AES128-SHA:DHE-RSA-AES256-SHA256:DHE-RSA-AES256-SHA" \
+    ":ECDHE-ECDSA-DES-CBC3-SHA:ECDHE-RSA-DES-CBC3-SHA:EDH-RSA-DES-CBC3-SHA" \
+    ":AES128-GCM-SHA256:AES256-GCM-SHA384:AES128-SHA256:AES256-SHA256" \
+    ":AES128-SHA:AES256-SHA:DES-CBC3-SHA" \
+    ":!aNULL:!eNULL:!EXPORT:!DES:!MD5:!PSK:!RC4:!DSS";
+
 /*! Global context for the sstp-client */
 static sstp_client_st client;
 
@@ -49,10 +69,9 @@ static void sstp_client_proxy_connected(sstp_stream_st *stream, sstp_buff_st *bu
 
 static void sstp_client_event_cb(sstp_client_st *client, int ret)
 {
-    uint8_t *skey;
-    uint8_t *rkey;
-    size_t   slen;
-    size_t   rlen;
+    uint8_t *skey = NULL;
+    uint8_t *rkey = NULL;
+    size_t   klen = 0;
 
     /* Check the result of the event */
     if (SSTP_OKAY != ret)
@@ -61,14 +80,14 @@ static void sstp_client_event_cb(sstp_client_st *client, int ret)
     }
 
     /* Get the result */
-    ret = sstp_event_mppe_result(client->event, &skey, &slen, &rkey, &rlen);
+    ret = sstp_event_mppe_result(client->event, &skey, &rkey, &klen);
     if (SSTP_OKAY != ret)
     {
         sstp_die("Failed to obtain the MPPE keys", -1);
     }
 
     /* Set the MPPE keys */
-    sstp_state_mppe_keys(client->state, skey, slen, rkey, rlen);
+    sstp_state_mppe_keys(client->state, skey, rkey, klen);
 
     /* Tell the state machine to connect */
     ret = sstp_state_accept(client->state);
@@ -115,7 +134,7 @@ static void sstp_client_pppd_cb(sstp_client_st *client, sstp_pppd_event_t ev)
         }
 
         /* Set the keys */
-        sstp_state_mppe_keys(client->state, skey, 16, rkey, 16);
+        sstp_state_mppe_keys(client->state, skey, rkey, 16);
         break;
     }
 
@@ -199,7 +218,8 @@ static void sstp_client_state_cb(sstp_client_st *client, sstp_state_t event)
  */
 static void sstp_client_http_done(sstp_client_st *client, int status)
 {
-    int opts = SSTP_VERIFY_NONE;
+    sstp_option_st *opts = &client->option;
+    int vopts = SSTP_VERIFY_NONE;
 
     if (SSTP_OKAY != status)
     {
@@ -211,15 +231,14 @@ static void sstp_client_http_done(sstp_client_st *client, int status)
     client->http = NULL;
 
     /* Set verify options */
-    opts = SSTP_VERIFY_NAME;
-    if (client->option.ca_cert ||
-        client->option.ca_path)
+    vopts = SSTP_VERIFY_NAME;
+    if (opts->ca_cert || opts->ca_path)
     {
-        opts = SSTP_VERIFY_CERT;
+        vopts = SSTP_VERIFY_CERT;
     }
 
     /* Verify the server certificate */
-    status = sstp_verify_cert(client->stream, client->option.server, opts);
+    status = sstp_verify_cert(client->stream, opts->host ?: opts->server, vopts);
     if (SSTP_OKAY != status)
     {
         if (!(SSTP_OPT_CERTWARN & client->option.enable))
@@ -302,18 +321,8 @@ static void sstp_client_proxy_done(sstp_client_st *client, int status)
         sstp_stream_destroy(client->stream);
 
         /* Create the SSL I/O streams */
-        if (SSTP_OPT_TLSEXT & client->option.enable)
-        {
-            log_info("TLS hostname extension is enabled");
-            ret = sstp_stream_create(&client->stream, client->ev_base,
-                    client->ssl_ctx, client->host.name);
-        }
-        else
-        {
-            log_info("TLS hostname extension is disabled");
-            ret = sstp_stream_create(&client->stream, client->ev_base,
-                    client->ssl_ctx, NULL);
-        }
+    	ret = sstp_stream_create(&client->stream, client->ev_base,
+	    	client->ssl_ctx, &client->option);
         if (SSTP_OKAY != ret)
         {
             sstp_die("Could not create I/O stream", -1);
@@ -335,15 +344,17 @@ static void sstp_client_proxy_done(sstp_client_st *client, int status)
         break;
 
     case SSTP_OKAY:
+    {
+        sstp_option_st *opts = &client->option;
 
         log_info("Connected to %s via proxy server", 
-                client->option.server);
+                opts->host ?: opts->server);
 
         /* Re-initialize the HTTP context */
         sstp_http_free(client->http);
 
         /* Create the HTTP handshake context */
-        ret = sstp_http_create(&client->http, client->option.server, (sstp_http_done_fn) 
+        ret = sstp_http_create(&client->http, opts->host ?: opts->server, (sstp_http_done_fn) 
                 sstp_client_http_done, client, SSTP_MODE_CLIENT);
         if (SSTP_OKAY != ret)
         {
@@ -358,7 +369,7 @@ static void sstp_client_proxy_done(sstp_client_st *client, int status)
         }
 
         break;
-
+    }
     default:
 
         sstp_die("Could not connect to proxy server", -1);
@@ -385,8 +396,8 @@ static void sstp_client_proxy_connected(sstp_stream_st *stream, sstp_buff_st *bu
     /* Create the HTTP object if one doesn't already exist */
     if (!client->http) 
     {
-        ret = sstp_http_create(&client->http, client->option.server,
-            (sstp_http_done_fn) sstp_client_proxy_done, client, SSTP_MODE_CLIENT);
+        ret = sstp_http_create(&client->http, client->option.host ?: client->option.server,
+                (sstp_http_done_fn) sstp_client_proxy_done, client, SSTP_MODE_CLIENT);
         if (SSTP_OKAY != ret)
         {
             sstp_die("Could not configure HTTP handshake with server", -1);
@@ -410,23 +421,27 @@ static void sstp_client_proxy_connected(sstp_stream_st *stream, sstp_buff_st *bu
 static status_t sstp_client_connect(sstp_client_st *client, 
         struct sockaddr *addr, int alen)
 {
-    sstp_client_cb complete_cb = (client->option.proxy)
-            ? sstp_client_proxy_connected
-            : sstp_client_connected;
+    sstp_option_st *opts = &client->option;
+    sstp_client_cb complete_cb = sstp_client_proxy_connected;
     status_t ret = SSTP_FAIL;
+ 
+    /* A likely condition */   
+    if (!opts->proxy) 
+    {
+        /* Resolved name is the same as specified, then server is an ip-address */
+        if (!strcmp(client->host.name, opts->server) &&
+             opts->host != NULL) 
+        {
+            /* Use the host per --host option, if specified */
+            strncpy(client->host.name, opts->host, 
+                    sizeof(client->host.name));
+        }
+        complete_cb = sstp_client_connected;
+    }
 
     /* Create the I/O streams */
-    if (SSTP_OPT_TLSEXT & client->option.enable)
-    {
-        log_info("TLS hostname extension is enabled");
-        ret = sstp_stream_create(&client->stream, client->ev_base, client->ssl_ctx, client->host.name);
-    }
-    else
-    {
-        log_info("TLS hostname extension is disabled");
-        ret = sstp_stream_create(&client->stream, client->ev_base, client->ssl_ctx, NULL);
-    }
-
+    ret = sstp_stream_create(&client->stream, client->ev_base, 
+            client->ssl_ctx, opts);
     if (SSTP_OKAY != ret)
     {
         log_err("Could not setup SSL streams");
@@ -434,7 +449,8 @@ static status_t sstp_client_connect(sstp_client_st *client,
     }
 
     /* Have the stream connect */
-    ret = sstp_stream_connect(client->stream, addr, alen, (sstp_complete_fn) complete_cb, client, 10);
+    ret = sstp_stream_connect(client->stream, addr, alen, (sstp_complete_fn)
+            complete_cb, client, 10);
     if (SSTP_INPROG != ret && 
         SSTP_OKAY   != ret)
     {
@@ -452,6 +468,72 @@ done:
 }
 
 
+/*!
+ * @brief Setup the CRL options
+ */
+static status_t sstp_setup_crl_options(SSL_CTX *ctx, sstp_option_st *opt)
+{
+    int status = SSTP_FAIL;
+
+    X509_STORE *certstore = SSL_CTX_get_cert_store(ctx);
+    if (!certstore)
+    {
+        log_err("Could not get certificate store");
+        goto done;
+    }
+
+    /* Configure the CRL directory if specified */
+    if (opt->crl_path)
+    {
+        X509_LOOKUP *lookup = X509_STORE_add_lookup(certstore, X509_LOOKUP_hash_dir());
+        if (!lookup)
+        {
+            log_err("Could get lookup structure");
+            goto done;
+        }
+
+        if (!X509_LOOKUP_add_dir(lookup, opt->crl_path, X509_FILETYPE_PEM))
+        {
+            log_err("Could not add directory");
+            goto done;
+        }
+    }
+
+    /* Configure the CRL file if specified */
+    if (opt->crl_file)
+    {
+        FILE *fp = fopen(opt->crl_file, "r");
+        if (!fp)
+        {
+            log_err("Cannot open CRL file '%s'", opt->crl_file);
+            goto done;
+        }
+
+        X509_CRL *crl = PEM_read_X509_CRL(fp, NULL, NULL, NULL);
+        if (!crl)
+        {
+            log_err("Cannot read CRL file '%s'", opt->crl_file);
+            goto done;
+        }
+
+        if (!X509_STORE_add_crl(certstore, crl))
+        {
+            log_err("Cannot add CRL to certificate store");
+            goto done;
+        }
+    }
+
+    X509_STORE_set_flags(certstore, X509_V_FLAG_CRL_CHECK);
+
+    /* Success! */
+    status = SSTP_OKAY;
+
+done:
+
+    return status;
+}
+
+
 /*!
  * @brief Perform the global SSL initializers
  */
@@ -479,19 +561,52 @@ static status_t sstp_init_ssl(sstp_client_st *client, sstp_option_st *opt)
         goto done;
     }
 
-    /* Configure the crypto options, eliminate SSLv2 */
-    status = SSL_CTX_set_options(client->ssl_ctx, SSL_OP_ALL|SSL_OP_NO_SSLv2);
+    /* Configure the crypto options, eliminate SSLv2, SSLv3 */
+    status = SSL_CTX_set_options(
+        client->ssl_ctx,
+        SSL_OP_ALL |
+            SSL_OP_NO_SSLv2 |
+            SSL_OP_NO_SSLv3);
     if (status == -1)
     {
         log_err("Could not set SSL options");
         goto done;
     }
 
+#ifdef SSL_OP_NO_COMPRESSION
+    /* disable to mitigate CRIME attack */
+    status = SSL_CTX_set_options(client->ssl_ctx, SSL_OP_NO_COMPRESSION);
+    if (status == -1)
+    {
+        log_err("Could not disable compression");
+        goto done;
+    }
+#endif
+
+#ifdef SSL_MODE_RELEASE_BUFFERS
+    /* reduce idle connection memory usage */
+    status = SSL_CTX_set_mode(client->ssl_ctx, SSL_MODE_RELEASE_BUFFERS);
+    if (status == -1)
+    {
+        log_err("Could not set option to optimize buffer usage");
+        goto done;
+    }
+#endif
+
+    status = SSL_CTX_set_cipher_list(
+        client->ssl_ctx,
+        sstp_client_ssl_ciphers);
+    if (status != 1)
+    {
+        log_err("Could not set SSL ciphersuites");
+        goto done;
+    }
+
     /* Configure the CA-Certificate or Directory */
     if (opt->ca_cert || opt->ca_path)
     {
         /* Look for certificates in the default certificate path */
-        status = SSL_CTX_load_verify_locations(client->ssl_ctx, 
+        status = SSL_CTX_load_verify_locations(client->ssl_ctx,
                 opt->ca_cert, opt->ca_path);
         if (status != 1)
         {
@@ -500,6 +615,17 @@ static status_t sstp_init_ssl(sstp_client_st *client, sstp_option_st *opt)
         }
     }
 
+    /* Configure the CRL file or directory */
+    if (opt->crl_file || opt->crl_path)
+    {
+        status = sstp_setup_crl_options(client->ssl_ctx, opt);
+        if (status != SSTP_OKAY)
+        {
+            log_err("Could not configure CRL options");
+            goto done;
+        }
+    }
+
     /* OBS: In case of longer certificate chains than 1 */
     SSL_CTX_set_verify_depth(client->ssl_ctx, 9);
 
@@ -798,14 +924,14 @@ int main(int argc, char *argv[])
         }
     }
 
-    /* Lookup the URL of the proxy server */
+    /* Lookup the URL of the server (or proxy if enabled) */
     ret = sstp_client_lookup(client.url, &client.host);
     if (SSTP_OKAY != ret)
     {
         sstp_die("Could not lookup host: `%s'", -1, client.url->host);
     }
-
-    /* Connect to the server */
+ 
+    /* Connect to the server (or proxy if enabled) */
     ret = sstp_client_connect(&client, &client.host.addr, 
             client.host.alen);
     if (SSTP_FAIL == ret)
